from degenbot.uniswap.abi import UNISWAP_V2_POOL_ABI as SUSHISWAP_V2_POOL_ABI
from degenbot.uniswap.abi import UNISWAP_V3_POOL_ABI as SUSHISWAP_V3_POOL_ABI

__all__ = (
    "SUSHISWAP_V2_POOL_ABI",
    "SUSHISWAP_V3_POOL_ABI",
)
