from degenbot import get_checksum_address

CURVE_V1_REGISTRY_ADDRESS = get_checksum_address("0x90E00ACe148ca3b23Ac1bC8C240C2a7Dd9c2d7f5")
CURVE_V1_FACTORY_ADDRESS = get_checksum_address("0x127db66E7F0b16470Bec194d0f496F9Fa065d0A9")
CURVE_V1_METAREGISTRY_ADDRESS = get_checksum_address("0xF98B45FA17DE75FB1aD0e7aFD971b0ca00e379fC")

BROKEN_CURVE_V1_POOLS = frozenset(
    get_checksum_address(pool_address)
    for pool_address in (
        "0x110cc323ca53d622469EdD217387E2E6B33F1dF5",
        "0x1F71f05CF491595652378Fe94B7820344A551B8E",
        "0x28B0Cf1baFB707F2c6826d10caf6DD901a6540C5",
        "0x3685646651FCcC80e7CCE7Ee24c5f47Ed9b434ac",
        "0x84997FAFC913f1613F51Bb0E2b5854222900514B",
        "0x9CA41a2DaB3CEE15308998868ca644e2e3be5C59",
        "0xA77d09743F77052950C4eb4e6547E9665299BecD",
        "0xD652c40fBb3f06d6B58Cb9aa9CFF063eE63d465D",
    )
)
