from degenbot.uniswap.v2_liquidity_pool import UniswapV2Pool
from degenbot.uniswap.v3_liquidity_pool import UniswapV3Pool


class SushiswapV2Pool(UniswapV2Pool): ...


class SushiswapV3Pool(UniswapV3Pool): ...
